# coding: utf-8
#


import pytest
import adbutils
from adb_server import encode


def test_encode():
    assert encode(1234) == b'000404d2'


def test_server_version(adb: adbutils.AdbClient):
    assert adb.server_version() == 1234


def test_server_kill(adb: adbutils.AdbClient):
    adb.server_kill()


def test_host_tport_serial(adb: adbutils.AdbClient):
    d = adb.device(serial="not-found")
    # pass
    with pytest.raises(adbutils.AdbError):
        d.open_transport()

    d = adb.device(serial="123456")
    d.open_transport()







# coding: utf-8
#

import logging
import threading
from unittest import mock
import adbutils
import pytest
import time
import socket
from adb_server import run_adb_server


def check_port(port) -> bool:
    try:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.settimeout(0.1)
            s.connect(('127.0.0.1', port))
        return True
    except (ConnectionRefusedError, OSError, socket.timeout):
        return False
    

def wait_for_port(port, timeout:float=3, ready: bool = True):
    start_time = time.time()
    while True:
        if time.time() - start_time > timeout:
            raise TimeoutError(f"Port {port} is not being listened to within {timeout} seconds")
        if check_port(port) == ready:
            return
        time.sleep(0.1)


@pytest.fixture(scope='function')
def adb_server_fixture():
    th = threading.Thread(target=run_adb_server, name='mock-adb-server')
    th.daemon = True
    th.start()
    wait_for_port(7305)
    yield
    adbutils.AdbClient(port=7305).server_kill()



@pytest.fixture
def adb(adb_server_fixture) -> adbutils.AdbClient:
    logging.basicConfig(level=logging.DEBUG)
    return adbutils.AdbClient(port=7305)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Sun Apr 07 2024 19:52:19 by codeskyblue
"""

import abc
import io
import logging
import re
from typing import Optional, Union
from adbutils.errors import AdbError
from adbutils.sync import Sync
from adbutils._proto import WindowSize
from PIL import Image

try:
    from PIL import UnidentifiedImageError
except ImportError:
    # fix for py37
    UnidentifiedImageError = OSError

logger = logging.getLogger(__name__)

class AbstractDevice(abc.ABC):
    @property
    @abc.abstractmethod
    def sync(self) -> Sync:
        pass

    @abc.abstractmethod
    def shell(self, cmd: str, encoding: Optional[str]) -> Union[str, bytes]:
        pass

    @abc.abstractmethod
    def window_size(self) -> WindowSize:
        pass

    @abc.abstractmethod
    def framebuffer(self) -> Image.Image:
        pass

class ScreenshotExtesion(AbstractDevice):
    def screenshot(self, display_id: Optional[int] = None, error_ok: bool = True) -> Image.Image:
        """ Take a screenshot and return PIL.Image.Image object
        Args:
            display_id: int, default None, see "dumpsys SurfaceFlinger --display-id" for valid display IDs
            error_ok: bool, default True, if True, return a black image when capture failed
        
        Returns:
            PIL.Image.Image object
        
        Raises:
            AdbError: when capture failed and error_ok is False
        """
        try:
            pil_image = self.__screencap(display_id)
            if pil_image.mode == "RGBA":
                pil_image = pil_image.convert("RGB")
            return pil_image
        except UnidentifiedImageError as e:
            logger.warning("screencap error: %s", e)
            if error_ok:
                wsize = self.window_size()
                return Image.new("RGB", wsize, (0, 0, 0))
            else:
                raise AdbError("screencap error") from e
    
    def __screencap(self, display_id: int = None) -> Image.Image:
        """ Take a screenshot and return PIL.Image.Image object
        """
        # framebuffer() is not stable, so here still use screencap
        cmdargs = ['screencap', '-p']
        if display_id is not None:
            _id = self.__get_real_display_id(display_id)
            cmdargs.extend(['-d', _id])
        png_bytes = self.shell(cmdargs, encoding=None)
        return Image.open(io.BytesIO(png_bytes))

    def __get_real_display_id(self, display_id: int) -> str:
        # adb shell dumpsys SurfaceFlinger --display-id
        # Display 4619827259835644672 (HWC display 0): port=0 pnpId=GGL displayName="EMU_display_0"
        output = self.shell("dumpsys SurfaceFlinger --display-id")
        _RE = re.compile(r"Display (\d+) ")
        ids = _RE.findall(output)
        if not ids:
            raise AdbError("No display found, debug with 'dumpsys SurfaceFlinger --display-id'")
        if display_id >= len(ids):
            raise AdbError("Invalid display_id", display_id)
        return ids[display_id]


"""
extra functions test

效果生效与否不好判定（例如屏幕亮暗），部分用例仅作冒烟测试
"""

import os
import io
import pathlib
import re
import time
import filecmp
import uuid

import pytest

import adbutils
from adbutils import AdbDevice, Network, BrightnessMode
from adbutils.errors import AdbSyncError


def test_shell(device: AdbDevice):
    for text in ("foo", "?", "&", "but    123"):
        output = device.shell(['echo', '-n', text])
        assert output == text

    output = device.shell("pwd", rstrip=False)
    assert output in ["/\n", "/\r\n"]


def test_shell_without_encoding(device: AdbDevice):
    output = device.shell("echo -n hello", encoding=None)
    assert output == b"hello"

    ret = device.shell2("echo -n hello", encoding=None)
    assert ret.output == b"hello"


def test_shell_stream(device: AdbDevice):
    c = device.shell(["echo", "-n", "hello world"], stream=True)
    output = c.read_until_close()
    assert output == "hello world"


def test_adb_shell_raise_timeout(device: AdbDevice):
    with pytest.raises(adbutils.AdbTimeout):
        device.shell("sleep 10", timeout=1.0)


def test_shell2(device: AdbDevice):
    cmd = "echo -n 'hello'; false"
    res = device.shell2(cmd)
    assert res.output == "hello"
    assert res.returncode == 1
    assert res.command == cmd


def test_get_xxx(device: AdbDevice):
    assert device.get_serialno()
    assert device.get_state() == "device"
    # adb connect device devpath is "unknown"
    # assert device.get_devpath().startswith("usb:")


def test_battery(device: AdbDevice):
    print(device.battery().level)


def test_keyevent(device: AdbDevice):
    # make sure no error raised
    device.keyevent(4)
    device.volume_up(2)
    device.volume_down(3)
    device.volume_mute()


def test_brightness(device: AdbDevice):
    current_brightness = device.brightness_value
    device.brightness_value = 100
    assert device.brightness_value == 100
    device.brightness_value = current_brightness

    current_mode = device.brightness_mode
    if current_mode == BrightnessMode.AUTO:
        device.brightness_mode = BrightnessMode.MANUAL
        assert device.brightness_mode == BrightnessMode.MANUAL
    elif current_mode == BrightnessMode.MANUAL:
        device.brightness_mode = BrightnessMode.AUTO
        assert device.brightness_mode == BrightnessMode.AUTO
    device.brightness_mode = current_mode


def test_switch_screen(device: AdbDevice):
    device.switch_screen(False)
    device.switch_screen(True)


def test_switch_airplane(device: AdbDevice):
    device.switch_airplane(True)
    device.switch_airplane(False)


def test_switch_wifi(device: AdbDevice):
    device.switch_wifi(False)
    device.switch_wifi(True)


def test_swipe(device: AdbDevice):
    device.swipe(100, 100, 400, 400)


def test_click(device: AdbDevice):
    device.click(100, 100)


def test_send_keys(device: AdbDevice):
    device.send_keys("1234")


def test_wlan_ip(device: AdbDevice):
    device.switch_airplane(False)
    device.switch_wifi(True)
    time.sleep(3)
    ip = device.wlan_ip()
    assert ip, 'ip is empty'


def test_app_start_stop(device: AdbDevice):
    d = device
    package_name = "io.appium.android.apis"
    if package_name not in d.list_packages():
        pytest.skip(package_name + " should be installed, to start test")
    d.app_start(package_name)
    time.sleep(1)
    assert device.app_current().package == package_name
    d.app_stop(package_name)
    time.sleep(.5)
    assert device.app_current().package != package_name


def test_sync_pull_push(device: AdbDevice, device_tmp_path, tmp_path: pathlib.Path):
    src = io.BytesIO(b"Hello 1")
    device.sync.push(src, device_tmp_path)
    assert b"Hello 1" == device.sync.read_bytes(device_tmp_path)

    device.sync.push(b"Hello 12", device_tmp_path)
    assert "Hello 12" == device.sync.read_text(device_tmp_path)

    target_path = tmp_path / "hi.txt"
    target_path.write_text("Hello Android")
    dst_path = tmp_path / "dst.txt"
    dst_path.unlink(missing_ok=True)

    device.sync.push(target_path, device_tmp_path)
    assert "Hello Android" == device.sync.read_text(device_tmp_path)
    device.sync.pull(device_tmp_path, dst_path)
    assert "Hello Android" == dst_path.read_text(encoding="utf-8")

    data = b""
    for chunk in device.sync.iter_content(device_tmp_path):
        data += chunk
    assert b"Hello Android" == data


def test_sync_pull_file_push(device: AdbDevice, device_tmp_path, tmp_path: pathlib.Path):
    src = io.BytesIO(b"Hello 1")
    device.sync.push(src, device_tmp_path)
    assert b"Hello 1" == device.sync.read_bytes(device_tmp_path)

    device.sync.push(b"Hello 12", device_tmp_path)
    assert "Hello 12" == device.sync.read_text(device_tmp_path)

    target_path = tmp_path / "hi.txt"
    target_path.write_text("Hello Android")
    dst_path = tmp_path / "dst.txt"
    dst_path.unlink(missing_ok=True)

    device.sync.push(target_path, device_tmp_path)
    assert "Hello Android" == device.sync.read_text(device_tmp_path)
    device.sync.pull_file(device_tmp_path, dst_path)
    assert "Hello Android" == dst_path.read_text(encoding="utf-8")

    data = b""
    for chunk in device.sync.iter_content(device_tmp_path):
        data += chunk
    assert b"Hello Android" == data


def test_sync_push_to_dir(device: AdbDevice, device_tmp_dir, tmp_path: pathlib.Path):
    random_data = str(uuid.uuid4()).encode()
    src = io.BytesIO(random_data)
    with pytest.raises(AdbSyncError):
        device.sync.push(src, device_tmp_dir)
    src_path = tmp_path.joinpath("random.txt")
    src_path.write_bytes(random_data)
    assert device.sync.push(src_path, device_tmp_dir) == len(random_data)
    assert random_data == device.sync.read_bytes(device_tmp_dir + "/random.txt")


def test_screenshot(device: AdbDevice):
    im = device.screenshot()
    assert im.mode == "RGB"


def test_framebuffer(device: AdbDevice):
    im = device.framebuffer()
    assert im.size


def test_app_info(device: AdbDevice):
    pinfo = device.app_current()
    app_info = device.app_info(pinfo.package)
    assert app_info.package_name is not None


def test_window_size(device: AdbDevice):
    w, h = device.window_size()
    assert isinstance(w, int)
    assert isinstance(h, int)


def test_is_screen_on(device: AdbDevice):
    bool_result = device.is_screen_on()
    assert isinstance(bool_result, bool)


def test_open_browser(device: AdbDevice):
    device.open_browser("https://example.org")


def test_dump_hierarchy(device: AdbDevice):
    output = device.dump_hierarchy()
    assert output.startswith("<?xml")
    assert output.rstrip().endswith("</hierarchy>")


def test_remove(device: AdbDevice):
    remove_path = "/data/local/tmp/touch.txt"
    device.shell(["touch", remove_path])
    assert device.sync.exists(remove_path)
    device.remove(remove_path)
    assert not device.sync.exists(remove_path)


# def test_create_connection(device: AdbDevice, device_tmp_path: str):
#     device.sync.push(b"hello", device_tmp_path)
#     device.create_connection(Network.LOCAL_FILESYSTEM, device_tmp_path)

def test_logcat(device: AdbDevice, tmp_path: pathlib.Path):
    logcat_path = tmp_path / "logcat.txt"
    logcat = device.logcat(logcat_path, clear=True, command="logcat -v time", re_filter="I/TAG")
    device.shell(["log", "-p", "i", "-t", "TAG", "hello"])
    time.sleep(.1)
    logcat.stop()
    assert logcat_path.exists()
    assert re.compile(r"I/TAG.*hello").search(logcat_path.read_text(encoding="utf-8"))


# todo: make independent of already present stuff on the phone
def test_pull_push_dirs(
        device: AdbDevice,
        device_tmp_dir_path: str,
        local_src_in_dir: pathlib.Path,
        tmp_path: pathlib.Path,
):
    def are_dir_trees_equal(dir1, dir2):
        """
        Compare two directories recursively. Files in each directory are
        assumed to be equal if their names and contents are equal.

        NB: retreived from: https://stackoverflow.com/a/6681395

        @param dir1: First directory path
        @param dir2: Second directory path

        @return: True if the directory trees are the same and 
            there were no errors while accessing the directories or files, 
            False otherwise.
        """

        dirs_cmp = filecmp.dircmp(dir1, dir2)
        if len(dirs_cmp.left_only) > 0 or len(dirs_cmp.right_only) > 0 or \
                len(dirs_cmp.funny_files) > 0:
            return False
        (_, mismatch, errors) = filecmp.cmpfiles(
            dir1, dir2, dirs_cmp.common_files, shallow=False)
        if len(mismatch) > 0 or len(errors) > 0:
            return False
        for common_dir in dirs_cmp.common_dirs:
            new_dir1 = os.path.join(dir1, common_dir)
            new_dir2 = os.path.join(dir2, common_dir)
            if not are_dir_trees_equal(new_dir1, new_dir2):
                return False
        return True

    local_src_out_dir1 = tmp_path / 'dir1'
    local_src_out_dir2 = tmp_path / 'dir2'

    # TODO: push src support dir
    # device.push(local_src_in_dir, device_tmp_dir_path)
    device.adb_output("push", str(local_src_in_dir), device_tmp_dir_path)

    device.sync.pull_dir(device_tmp_dir_path, local_src_out_dir1)

    assert local_src_out_dir1.exists()
    assert local_src_out_dir1.is_dir()

    are_dir_trees_equal(local_src_in_dir, local_src_out_dir1)

    device.sync.pull(device_tmp_dir_path, local_src_out_dir2)

    assert local_src_out_dir2.exists()
    assert local_src_out_dir2.is_dir()

    are_dir_trees_equal(local_src_in_dir, local_src_out_dir2)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Sun May 15 2022 22:12:45 by codeskyblue
"""

import os
import time
import adbutils


def test_screenrecord(device: adbutils.AdbDevice):
    device.start_recording("output.mp4")
    time.sleep(2.0)
    device.stop_recording()
    assert os.path.exists("output.mp4")
    os.remove("output.mp4")


# coding: utf-8

import pytest
from adbutils import AdbClient, adb, AdbDevice


@pytest.fixture(scope="session")
def device():
    client = AdbClient() # port=5137
    return client.device()


@pytest.fixture
def device_tmp_path(device: AdbDevice):
    tmp_path = "/data/local/tmp/Hi-世界.txt"
    yield tmp_path
    device.remove(tmp_path)


@pytest.fixture
def device_tmp_dir(device: AdbDevice):
    tmp_path = "/data/local/tmp/adbutils-test"
    device.shell("mkdir -p {}".format(tmp_path))
    yield tmp_path
    device.rmtree(tmp_path)


@pytest.fixture
def device_tmp_dir_path(device: AdbDevice):
    tmp_dir_path = "/sdcard/test_d"
    yield tmp_dir_path
    device.rmtree(tmp_dir_path)

@pytest.fixture
def local_src_in_dir(tmpdir):
    tmpdir.join('1.txt').write('1\n')
    tmpdir.join('2.txt').write('2\n')
    tmpdir.join('3.txt').write('3\n')

    a = tmpdir.mkdir('a')
    a.join('a1.txt').write('a1\n')

    aa = a.mkdir('aa')
    aa.join('aa1.txt').write('aa1\n')

    ab = a.mkdir('ab')
    ab.join('ab1.txt').write('ab1\n')
    ab.join('ab2.txt').write('ab2\n')

    b = tmpdir.mkdir('b')
    b.join('b1.txt').write('b1\n')

    c = tmpdir.mkdir('c')
    ca = c.mkdir('ca')
    ca.join('ca1.txt').write('ca1\n')

    caa = ca.mkdir('caa')
    caa.join('caa1.txt').write('caa1\n')

    cb = c.mkdir('cb')

    yield tmpdir



#
# Created on Sun Mar 24 2024 codeskyblue
#

from __future__ import annotations

import asyncio
import functools
import logging
import re
from typing import Union, overload

logger = logging.getLogger(__name__)


@overload
def encode(data: str) -> bytes:
    ...

@overload
def encode(data: bytes) -> bytes:
    ...

@overload
def encode(data: int) -> bytes:
    ...


def encode(data):
    if isinstance(data, bytes):
        return encode_bytes(data)
    if isinstance(data, int):
        return encode_number(data)
    if isinstance(data, str):
        return encode_string(data)
    raise ValueError("data must be bytes or int")


def encode_number(n: int) -> bytes:
    body = "{:04x}".format(n)
    return encode_bytes(body.encode())

def encode_string(s: str, encoding: str = 'utf-8') -> bytes:
    return encode_bytes(s.encode(encoding))

def encode_bytes(s: bytes) -> bytes:
    header = "{:04x}".format(len(s)).encode()
    return header + s



COMMANDS: dict[Union[str, re.Pattern], callable] = {}

def register_command(name: str):
    def wrapper(func):
        COMMANDS[name] = func
        return func
    return wrapper


class Context:
    def __init__(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter, server: "AdbServer" = None, command: str = None):
        self.reader = reader
        self.writer = writer
        self.server = server
        self.command = command
            
    async def send(self, data: bytes):
        self.writer.write(data)
        await self.writer.drain()

    async def recv(self, length: int) -> bytes:
        return await self.reader.read(length)

    async def recv_string_block(self) -> str:
        length = int((await self.recv(4)).decode(), 16)
        return (await self.recv(length)).decode()
    
    async def close(self):
        self.writer.close()
        await self.writer.wait_closed()


@register_command("host:version")
async def host_version(ctx: Context):
    await ctx.send(b"OKAY")
    await ctx.send(encode_number(1234))


@register_command("host:kill")
async def host_kill(ctx: Context):
    await ctx.send(b"OKAY")
    await ctx.close()
    await ctx.server.stop()
    # os.kill(os.getpid(), signal.SIGINT)

@register_command("host:list-forward")
async def host_list_forward(ctx: Context):
    await ctx.send(b"OKAY")
    await ctx.send(encode_string("123456 tcp:1234 tcp:4321"))


SHELL_OUTPUTS = {
    "pwd": "/",
}

@register_command(re.compile("host:tport:serial:.*"))
async def host_tport_serial(ctx: Context):
    serial = ctx.command.split(":")[-1]
    if serial == "not-found":
        await ctx.send(b"FAIL")
        await ctx.send(encode("device not found"))
    else:
        await ctx.send(b"OKAY")
        await ctx.send(b"\x00\x00\x00\x00\x00\x00\x00\x00")

    cmd = await ctx.recv_string_block()
    if not cmd.startswith("shell:"):
        await ctx.send(b"FAIL")
        await ctx.send(encode("unsupported command"))
        return
    await ctx.send(b"OKAY")
    shell_cmd = cmd.split(":", 1)[1]
    if shell_cmd in SHELL_OUTPUTS:
        await ctx.send((SHELL_OUTPUTS[shell_cmd].rstrip() + "\n").encode())
    else:
        await ctx.send(b"unknown command")
    


async def handle_command(reader: asyncio.StreamReader, writer: asyncio.StreamWriter, server: "AdbServer"):
    try:
         # Receive the command from the client
        addr = writer.get_extra_info('peername')
        logger.info(f"Connection from %s", addr)
        cmd_length = int((await reader.readexactly(4)).decode(), 16)
        command = (await reader.read(cmd_length)).decode()
        logger.info("recv command: %s", command)
        command_handler: callable = None
        command_keys = list(COMMANDS.keys())
        logger.debug("command_keys: %s", command_keys)
        for key in command_keys:
            if isinstance(key, str) and key == command:
                command_handler = COMMANDS[key]
                break
            elif isinstance(key, re.Pattern) and key.match(command):
                command_handler = COMMANDS[key]
                break

        logger.debug("command_handler: %s", command_handler)
        if command_handler is None:
            writer.write(b"FAIL")
            writer.write(encode(f"Unknown command: {command}"))
            await writer.drain()
            writer.close()
            return
        ctx = Context(reader, writer, server, command)
        await command_handler(ctx)
        await ctx.close()
    except asyncio.IncompleteReadError:
        pass



class AdbServer:
    def __init__(self, port: int = 7305, host: str = '127.0.0.1'):
        self.port = port
        self.host = host
        self.server = None
    
    async def start(self):
        _handle = functools.partial(handle_command, server=self)
        self.server = await asyncio.start_server(_handle, self.host, self.port)
        addr = self.server.sockets[0].getsockname()
        print(f'ADB server listening on {addr}')

        async with self.server:
            try:
                # Keep running the server
                await self.server.serve_forever()
            except asyncio.CancelledError:
                pass
    
    async def stop(self):
        self.server.close()
        await self.server.wait_closed()



async def adb_server():
    host = '127.0.0.1'
    port = 7305
    await AdbServer(port, host).start()


def run_adb_server():
    try:
        import logzero
        logzero.setup_logger(__name__)
    except ImportError:
        pass
    
    try:
        asyncio.run(adb_server())
    except:
        logger.exception("Error in adb_server")


if __name__ == '__main__':
    run_adb_server()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Created on Fri May 06 2022 10:33:39 by codeskyblue
"""

from __future__ import annotations

import pathlib
import re
import socket
import subprocess
import threading
import typing
from typing import List, Optional, Union

from PIL import Image, UnidentifiedImageError
from deprecation import deprecated

from adbutils._deprecated import DeprecatedExtension
from adbutils.install import InstallExtension
from adbutils.screenrecord import ScreenrecordExtension
from adbutils.screenshot import ScreenshotExtesion

from adbutils._adb import AdbConnection, BaseClient
from adbutils._proto import *
from adbutils._proto import StrOrPathLike
from adbutils._utils import StopEvent, adb_path, get_free_port, list2cmdline
from adbutils._version import __version__
from adbutils.errors import AdbError
from adbutils.shell import ShellExtension
from adbutils.sync import Sync


_DEFAULT_SOCKET_TIMEOUT = 600  # 10 minutes


class BaseDevice:
    """Basic operation for a device"""

    def __init__(
        self, client: BaseClient, serial: str = None, transport_id: int = None
    ):
        """
        Args:
            client (BaseClient): AdbClient instance
            serial (str): device serial
            transport_id (int): transport_id
        """
        self._client = client
        self._serial = serial
        self._transport_id: int = transport_id
        self._properties = {}  # store properties data

        if not serial and not transport_id:
            raise AdbError("serial, transport_id must set atleast one")

        self._prepare()

    def _prepare(self):
        """rewrite in sub class"""

    @property
    def serial(self) -> str:
        return self._serial

    def open_transport(
        self, command: str = None, timeout: float = _DEFAULT_SOCKET_TIMEOUT
    ) -> AdbConnection:
        # connect has it own timeout
        c = self._client.make_connection(timeout=timeout)

        if command:
            if self._transport_id:
                c.send_command(f"host-transport-id:{self._transport_id}:{command}")
                c.check_okay()
            elif self._serial:
                c.send_command(f"host-serial:{self._serial}:{command}")
                c.check_okay()
            else:
                raise RuntimeError("should not reach here")
        else:
            if self._transport_id:
                c.send_command(f"host:transport-id:{self._transport_id}")
                c.check_okay()
            elif self._serial:
                # host:tport:serial:xxx is also fine, but receive 12 bytes
                # recv: 4f 4b 41 59 14 00 00 00 00 00 00 00              OKAY........
                c.send_command(f"host:tport:serial:{self._serial}")
                c.check_okay()
                c.read(8)  # skip 8 bytes
            else:
                raise RuntimeError("should not reach here")
        return c

    def _get_with_command(self, cmd: str) -> str:
        c = self.open_transport(cmd)
        return c.read_string_block()

    def get_state(self) -> str:
        """return device state {offline,bootloader,device}"""
        return self._get_with_command("get-state")

    def get_serialno(self) -> str:
        """return the real device id, not the connect serial"""
        return self._get_with_command("get-serialno")

    def get_devpath(self) -> str:
        """example return: usb:12345678Y"""
        return self._get_with_command("get-devpath")

    def get_features(self) -> str:
        """
        Return example:
            'abb_exec,fixed_push_symlink_timestamp,abb,stat_v2,apex,shell_v2,fixed_push_mkdir,cmd'
        """
        return self._get_with_command("features")

    @property
    def info(self) -> dict:
        return {
            "serialno": self.get_serialno(),
            "devpath": self.get_devpath(),
            "state": self.get_state(),
        }

    def __repr__(self):
        return "AdbDevice(serial={})".format(self.serial)

    @property
    def sync(self) -> Sync:
        return Sync(self._client, self.serial)

    @property
    def prop(self) -> "Property":
        return Property(self)

    def adb_output(self, *args, **kwargs):
        """Run adb command use subprocess and get its content

        Returns:
            string of output

        Raises:
            EnvironmentError
        """
        cmds = [adb_path(), "-s", self._serial] if self._serial else [adb_path()]
        cmds.extend(args)
        try:
            return subprocess.check_output(
                cmds, stdin=subprocess.DEVNULL, stderr=subprocess.STDOUT
            ).decode("utf-8")
        except subprocess.CalledProcessError as e:
            if kwargs.get("raise_error", True):
                raise EnvironmentError(
                    "subprocess", cmds, e.output.decode("utf-8", errors="ignore")
                )

    def shell(
        self,
        cmdargs: Union[str, list, tuple],
        stream: bool = False,
        timeout: Optional[float] = _DEFAULT_SOCKET_TIMEOUT,
        encoding: str | None = "utf-8",
        rstrip=True,
    ) -> typing.Union[AdbConnection, str, bytes]:
        """Run shell inside device and get it's content

        Args:
            rstrip (bool): strip the last empty line (Default: True)
            stream (bool): return stream instead of string output (Default: False)
            timeout (float): set shell timeout
            encoding (str): set output encoding (Default: utf-8), set None to make return bytes
            rstrip (bool): strip the last empty line, only work when encoding is set

        Returns:
            string of output when stream is False
            AdbConnection when stream is True

        Raises:
            AdbTimeout

        Examples:
            shell("ls -l")
            shell(["ls", "-l"])
            shell("ls | grep data")
        """
        if isinstance(cmdargs, (list, tuple)):
            cmdargs = list2cmdline(cmdargs)
        if stream:
            timeout = None
        c = self.open_transport(timeout=timeout)
        c.send_command("shell:" + cmdargs)
        c.check_okay()
        if stream:
            return c
        output = c.read_until_close(encoding=encoding)
        if encoding:
            return output.rstrip() if rstrip else output
        return output

    def shell2(
        self,
        cmdargs: Union[str, list, tuple],
        timeout: Optional[float] = _DEFAULT_SOCKET_TIMEOUT,
        encoding: str | None = "utf-8",
        rstrip=False,
    ) -> ShellReturn:
        """
        Run shell command with detail output
        Args:
            cmdargs (str | list | tuple): command args
            timeout (float): set shell timeout, seconds
            encoding (str): set output encoding (Default: utf-8), set None to make return bytes
            rstrip (bool): strip the last empty line, only work when encoding is set

        Returns:
            ShellOutput

        Raises:
            AdbTimeout
        """
        if isinstance(cmdargs, (list, tuple)):
            cmdargs = list2cmdline(cmdargs)
        assert isinstance(cmdargs, str)
        MAGIC = "X4EXIT:"
        newcmd = cmdargs + f"; echo {MAGIC}$?"
        output = self.shell(newcmd, timeout=timeout, encoding=encoding, rstrip=True)
        rindex = output.rfind(MAGIC if encoding else MAGIC.encode())
        if rindex == -1:  # normally will not possible
            raise AdbError("shell output invalid", newcmd, output)
        returncoode = int(output[rindex + len(MAGIC) :])
        output = output[:rindex]
        if rstrip and encoding:
            output = output.rstrip()
        return ShellReturn(command=cmdargs, returncode=returncoode, output=output)

    def forward(self, local: str, remote: str, norebind: bool = False):
        self._client.forward(self._serial, local, remote, norebind)

    def forward_port(self, remote: Union[int, str]) -> int:
        """forward remote port to local random port"""
        if isinstance(remote, int):
            remote = "tcp:" + str(remote)
        for f in self.forward_list():
            if (
                f.serial == self._serial
                and f.remote == remote
                and f.local.startswith("tcp:")
            ):  # yapf: disable
                return int(f.local[len("tcp:") :])
        local_port = get_free_port()
        self.forward("tcp:" + str(local_port), remote)
        return local_port

    def forward_list(self) -> List[ForwardItem]:
        items = self._client.forward_list()
        return [item for item in items if item.serial == self._serial]

    def reverse(self, remote: str, local: str, norebind: bool = False):
        """
        Args:
            serial (str): device serial
            remote, local (str):
                - tcp:<port>
                - localabstract:<unix domain socket name>
                - localreserved:<unix domain socket name>
                - localfilesystem:<unix domain socket name>
            norebind (bool): fail if already reversed when set to true

        Raises:
            AdbError
        """
        c = self.open_transport()
        args = ["reverse:forward"]
        if norebind:
            args.append("norebind")
        args.append(remote + ";" + local)
        c.send_command(":".join(args))
        c.check_okay() # this OKAY means message was received
        c.check_okay() # check reponse

    def reverse_list(self) -> List[ReverseItem]:
        c = self.open_transport()
        c.send_command("reverse:list-forward")
        c.check_okay()
        content = c.read_string_block()
        items = []
        for line in content.splitlines():
            parts = line.split()
            if len(parts) != 3:
                continue
            items.append(ReverseItem(*parts[1:]))
        return items
    
    def framebuffer(self) -> Image.Image:
        """Capture device screen and return PIL.Image object (Not very stable)

        Raises:
            NotImplementedError
        """
        # Ref: https://android.googlesource.com/platform/system/core/+/android-cts-7.0_r18/adb/framebuffer_service.cpp
        # Ref: https://github.com/DeviceFarmer/adbkit/blob/c16081384ca34addbdab318bda3c76434b7538af/src/adb/command/host-transport/framebuffer.ts
        c = self.open_transport()
        c.send_command("framebuffer:")
        c.check_okay()
        
        version = c.read_uint32()
        if version == 16:
            raise NotImplementedError("Unsupported version 16")
        bpp = c.read_uint32() # bits per pixel
        if bpp != 24 and bpp != 32:
            raise NotImplementedError("Unsupported bpp(bits per pixel)", bpp)
        size = c.read_uint32()
        if size == 1:
            # FIXME: what is this?
            size = c.read_uint32()
        width = c.read_uint32()
        height = c.read_uint32()
        red_offset = c.read_uint32()
        red_length = c.read_uint32() # always 8
        blue_offset = c.read_uint32()
        blue_length = c.read_uint32() # always 8
        green_offset = c.read_uint32()
        green_length = c.read_uint32() # always 8
        alpha_offset = c.read_uint32()
        alpha_length = c.read_uint32()

        color_format = 'RGB'
        if blue_offset == 0:
            color_format = 'BGR'
        if bpp == 32 or alpha_length:
            color_format += 'A'

        if color_format != 'RGBA' and color_format != 'RGB':
            raise NotImplementedError("Unsupported color format")
        buffer = c.read(size)
        if len(buffer) != size:
            raise UnidentifiedImageError("framebuffer size not match", size, len(buffer))
        image = Image.frombytes(color_format, (width, height), buffer)
        return image

    @deprecated(deprecated_in="2.6.0", removed_in="3.0.0", current_version=__version__, details="use sync.push instead")
    def push(self, local: str, remote: str):
        """ alias for sync.push """
        return self.sync.push(local, remote)

    def create_connection(
        self, network: Network, address: Union[int, str]
    ) -> socket.socket:
        """
        Used to connect a socket (unix of tcp) on the device

        Returns:
            socket object

        Raises:
            AssertionError, ValueError
        """
        c = self.open_transport()
        if network == Network.TCP:
            assert isinstance(address, int)
            c.send_command("tcp:" + str(address))
            c.check_okay()
        elif network in [Network.UNIX, Network.LOCAL_ABSTRACT]:
            assert isinstance(address, str)
            c.send_command("localabstract:" + address)
            c.check_okay()
        elif network in [
            Network.LOCAL_FILESYSTEM,
            Network.LOCAL,
            Network.DEV,
            Network.LOCAL_RESERVED,
        ]:
            c.send_command(network + ":" + address)
            c.check_okay()
        else:
            raise ValueError("Unsupported network type", network)
        c._finalizer.detach()
        return c.conn

    def root(self):
        """restart adbd as root

        Return example:
            cannot run as root in production builds
        """
        # Ref: https://github.com/Swind/pure-python-adb/blob/master/ppadb/command/transport/__init__.py#L179
        c = self.open_transport()
        c.send_command("root:")
        c.check_okay()
        return c.read_until_close()

    def tcpip(self, port: int):
        """restart adbd listening on TCP on PORT

        Return example:
            restarting in TCP mode port: 5555
        """
        c = self.open_transport()
        c.send_command("tcpip:" + str(port))
        c.check_okay()
        return c.read_until_close()

    def logcat(
        self,
        file: StrOrPathLike = None,
        clear: bool = False,
        re_filter: typing.Union[str, re.Pattern] = None,
        command: str = "logcat -v time",
    ) -> StopEvent:
        """
        Args:
            file (str): file path to save logcat
            clear (bool): clear logcat before start
            re_filter (str | re.Pattern): regex pattern to filter logcat
            command (str): logcat command, default is "logcat -v time"

        Example usage:
            >>> evt = device.logcat("logcat.txt", clear=True, re_filter=".*python.*")
            >>> time.sleep(10)
            >>> evt.stop()
        """
        if re_filter:
            if isinstance(re_filter, str):
                re_filter = re.compile(re_filter)
            assert isinstance(re_filter, re.Pattern)

        if clear:
            self.shell("logcat --clear")

        def _filter_func(line: str) -> bool:
            if re_filter is None:
                return True
            return re_filter.search(line) is not None

        def _copy2file(
            stream: AdbConnection,
            fdst: typing.TextIO,
            event: StopEvent,
            filter_func: typing.Callable[[str], bool],
        ):
            try:
                fsrc = stream.conn.makefile("r", encoding="UTF-8", errors="replace")
                while not event.is_stopped():
                    line = fsrc.readline()
                    if not line:
                        break
                    if filter_func(line):
                        fdst.write(line)
                        fdst.flush()
            finally:
                fsrc.close()
                stream.close()
                event.done()

        event = StopEvent()
        stream = self.shell(command, stream=True)
        fdst = pathlib.Path(file).open("w", encoding="UTF-8")
        threading.Thread(
            name="logcat",
            target=_copy2file,
            args=(stream, fdst, event, _filter_func),
            daemon=True,
        ).start()
        return event


class Property:

    def __init__(self, d: BaseDevice):
        self._d = d

    def __str__(self):
        return f"product:{self.name} model:{self.model} device:{self.device}"

    def get(self, name: str, cache=True) -> str:
        if cache and name in self._d._properties:
            return self._d._properties[name]
        value = self._d._properties[name] = self._d.shell(["getprop", name]).strip()
        return value

    @property
    def name(self):
        return self.get("ro.product.name", cache=True)

    @property
    def model(self):
        return self.get("ro.product.model", cache=True)

    @property
    def device(self):
        return self.get("ro.product.device", cache=True)


class AdbDevice(
    BaseDevice,
    ShellExtension,
    ScreenrecordExtension,
    ScreenshotExtesion,
    InstallExtension,
    DeprecatedExtension,
):
    """provide custom functions for some complex operations"""

    def __init__(
        self, client: BaseClient, serial: str = None, transport_id: int = None
    ):
        BaseDevice.__init__(self, client, serial, transport_id)
        ScreenrecordExtension.__init__(self)



#!/usr/bin/env python3
# coding: utf-8
import os
import shutil
import subprocess
import sys
import zipfile

import requests

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
LIBNAME = "adbutils"
BINARIES_DIR = os.path.join(ROOT_DIR, LIBNAME, "binaries")

FNAMES_PER_PLATFORM = {
    "darwin": ["adb"],
    "linux": ["adb"],
    "win32": ["adb.exe", "AdbWinApi.dll", "AdbWinUsbApi.dll"],
}

BINARIES_URL = {
    "darwin": "https://dl.google.com/android/repository/platform-tools-latest-darwin.zip",
    "linux": "https://dl.google.com/android/repository/platform-tools-latest-linux.zip",
    "win32": "https://dl.google.com/android/repository/platform-tools-latest-windows.zip"
}

# https://peps.python.org/pep-0491/#file-format
# https://packaging.python.org/en/latest/specifications/platform-compatibility-tags/
linux_plats = "manylinux1_x86_64"
darwin_plats = "macosx_10_9_intel.macosx_10_9_x86_64.macosx_10_10_intel.macosx_10_10_x86_64"

WHEEL_BUILDS = {
    "py3-none-" + darwin_plats: "darwin",
    "py3-none-" + linux_plats: "linux",  # look into manylinux wheel builder
    "py3-none-win32": "win32",
    "py3-none-win_amd64": "win32",
}


def copy_binaries(target_dir, platform: str):
    assert os.path.isdir(target_dir)

    base_url = BINARIES_URL[platform]
    archive_name = os.path.join(target_dir, f'{platform}.zip')

    print("Downloading", base_url, "...", end=" ", flush=True)
    with open(archive_name, 'wb') as handle:
        response = requests.get(base_url, stream=True)
        if not response.ok:
            print(response)
        for block in response.iter_content(1024):
            if not block:
                break
            handle.write(block)
    print("done")

    for fname in FNAMES_PER_PLATFORM[platform]:
        print("Extracting", fname, "...", end=" ")
        # extract the specified file from the archive
        member_name = f'platform-tools/{fname}'
        extract_archive_file(archive_file=archive_name, file=member_name, destination_folder=target_dir)
        shutil.move(src=os.path.join(target_dir, member_name), dst=os.path.join(target_dir, fname))

        # extracted files
        filename = os.path.join(target_dir, fname)
        if fname == "adb":
            os.chmod(filename, 0o755)
        print("done")

    os.rmdir(path=os.path.join(target_dir, 'platform-tools'))
    os.remove(path=archive_name)


def extract_archive_file(archive_file, file, destination_folder):
    extension = archive_file.rsplit('.', 1)[-1].lower()

    if extension == 'zip':
        with zipfile.ZipFile(archive_file, 'r') as archive:
            archive.extract(member=file, path=destination_folder)


def clear_binaries_dir(target_dir):
    assert os.path.isdir(target_dir)
    assert os.path.basename(target_dir) == "binaries"
    
    for fname in os.listdir(target_dir):
        if fname == "README.md" or fname.endswith(".py"):
            continue
        print("Removing", fname, "...", end=" ")
        os.remove(os.path.join(target_dir, fname))
        print("done")


def clean():
    for root, dirs, files in os.walk(ROOT_DIR):
        for dname in dirs:
            if dname in (
                "__pycache__",
                ".cache",
                "dist",
                "build",
                LIBNAME + ".egg-info",
            ):
                shutil.rmtree(os.path.join(root, dname))
                print("Removing", dname)
        for fname in files:
            if fname.endswith((".pyc", ".pyo")):
                os.remove(os.path.join(root, fname))
                print("Removing", fname)


def build():
    clean()
    # Clear binaries, we don't want them in the reference release
    clear_binaries_dir(BINARIES_DIR)
    
    print("Using setup.py to generate wheel ...", end="")
    subprocess.check_output(
        [sys.executable, "setup.py", "sdist", "bdist_wheel"], cwd=ROOT_DIR
    )
    print("done")
    
    # Version is generated by pbr
    version = None
    distdir = os.path.join(ROOT_DIR, "dist")
    for fname in os.listdir(distdir):
        if fname.endswith(".whl"):
            version = fname.split("-")[1]
            break
    assert version

    # Prepare
    fname = "-".join([LIBNAME, version, "py3-none-any.whl"])
    packdir = LIBNAME+"-"+version
    infodir = f"{LIBNAME}-{version}.dist-info"
    wheelfile = os.path.join(distdir, packdir, infodir, "WHEEL")
    print("Path:", os.path.join(distdir, fname))
    assert os.path.isfile(os.path.join(distdir, fname))
    
    print("Unpacking ...", end="")
    subprocess.check_output(
        [sys.executable, "-m", "wheel", "unpack", fname], cwd=distdir)
    os.remove(os.path.join(distdir, packdir, infodir, "RECORD"))
    print("done")
    
    # Build for different platforms
    for wheeltag, platform in WHEEL_BUILDS.items():
        print(f"Edit for {platform} {wheeltag}")

        # copy binaries
        binary_dir = os.path.join(distdir, packdir, LIBNAME, "binaries")
        clear_binaries_dir(binary_dir)
        copy_binaries(binary_dir, platform)
        
        lines = []
        for line in open(wheelfile, "r", encoding="UTF-8"):
            if line.startswith("Tag:"):
                line = "Tag: " + wheeltag
            lines.append(line.rstrip())
        with open(wheelfile, "w", encoding="UTF-8") as f:
            f.write("\n".join(lines))
        
        print("Pack ...", end="")
        subprocess.check_output(
            [sys.executable, "-m", "wheel", "pack", packdir], cwd=distdir)
        print("done")
    
    # Clean up
    os.remove(os.path.join(distdir, fname))
    shutil.rmtree(os.path.join(distdir, packdir))
    
    # Show overview
    print("Dist folder:")
    for fname in sorted(os.listdir(distdir)):
        s = os.stat(os.path.join(distdir, fname)).st_size
        print("  {:0.0f} KB {}".format(s / 2**10, fname))


def release():
    """ Release the packages to pypi """
    username = os.environ["PYPI_USERNAME"]
    password = os.environ['PYPI_PASSWORD']
    subprocess.check_call(
        [sys.executable, "-m", "twine", "upload", "-u", username, '-p', password, "dist/*"])


if __name__ == "__main__":
    build()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Wed May 08 2024 21:45:15 by codeskyblue
"""

import adbutils


def test_forward_list(adb: adbutils.AdbClient):
    items = adb.forward_list()
    assert len(items) == 1
    assert items[0].serial == "123456"
    assert items[0].local == "tcp:1234"
    assert items[0].remote == "tcp:4321"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Mon May 06 2024 14:41:10 by codeskyblue
"""

from unittest import mock
import pytest
import adbutils
from adbutils.errors import AdbError


def test_shell_pwd(adb: adbutils.AdbClient):
    d = adb.device(serial="123456")
    assert d.shell("pwd") == "/"


def test_shell_screenshot(adb: adbutils.AdbClient):
    d = adb.device(serial="123456")
    
    def mock_shell(cmd: str, encoding='utf-8', **kwargs):
        if encoding is None:
            return b""
        if cmd == "wm size":
            return "Physical size: 1080x1920"
        return b""

    d.shell = mock_shell
    d.rotation = lambda: 0

    with pytest.raises(AdbError):
        d.screenshot(error_ok=False)
    pil_img = d.screenshot(error_ok=True)
    assert pil_img.size == (1080, 1920)
    
    # assert pixel is blank
    pixel = pil_img.getpixel((0, 0))
    assert pixel[:3] == (0, 0, 0)


def test_window_size(adb: adbutils.AdbClient):
    d = adb.device(serial="123456")
    
    def mock_shell(cmd):
        if cmd == "wm size":
            return "Physical size: 1080x1920"
        if cmd == "dumpsys display":
            return "mViewports=[DisplayViewport{orientation=0]"
        return ""
    
    d.shell = mock_shell
    wsize = d.window_size()
    assert wsize.width == 1080
    assert wsize.height == 1920


def test_shell_battery(adb: adbutils.AdbClient):
    d = adb.device(serial="123456")

    _DUMPSYS_BATTERY_ = """Current Battery Service state:
    AC powered: false
    USB powered: true
    Wireless powered: false
    Dock powered: false
    Max charging current: 0
    Max charging voltage: 0
    Charge counter: 10000
    status: 4
    health: 2
    present: true
    level: 80
    scale: 100
    voltage: 5000
    temperature: 250
    technology: Li-ion"""
    d.shell = lambda cmd: _DUMPSYS_BATTERY_
    
    bat = d.battery()
    assert bat.ac_powered == False
    assert bat.wireless_powered == False
    assert bat.usb_powered == True
    assert bat.dock_powered == False
    assert bat.max_charging_current == 0
    assert bat.max_charging_voltage == 0
    assert bat.charge_counter == 10000
    assert bat.status == 4
    assert bat.health == 2
    assert bat.present == True
    assert bat.level == 80
    assert bat.scale == 100
    assert bat.voltage == 5000
    assert bat.temperature == 25.0
    assert bat.technology == "Li-ion"


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Created on Fri May 06 2022 11:39:40 by codeskyblue
"""
from __future__ import annotations

__all__ = [
    "Network", "BrightnessMode", "DeviceEvent", "ForwardItem", "ReverseItem", "FileInfo",
    "WindowSize", "RunningAppInfo", "ShellReturn", "AdbDeviceInfo", "AppInfo", "BatteryInfo"
]

import enum
import datetime
import pathlib
from typing import List, NamedTuple, Optional, Union
from dataclasses import dataclass


class Network(str, enum.Enum):
    TCP = "tcp"
    UNIX = "unix"

    DEV = "dev"
    LOCAL = "local"
    LOCAL_RESERVED = "localreserved"
    LOCAL_FILESYSTEM = "localfilesystem"
    LOCAL_ABSTRACT = "localabstract"  # same as UNIX


class BrightnessMode(int, enum.Enum):
    AUTO = 1
    MANUAL = 0


@dataclass(frozen=True)
class DeviceEvent:
    present: bool
    serial: str
    status: str


@dataclass
class ForwardItem:
    serial: str
    local: str
    remote: str


@dataclass
class ReverseItem:
    remote: str
    local: str


@dataclass
class FileInfo:
    mode: int
    size: int
    mtime: datetime.datetime
    path: str


@dataclass
class AppInfo:
    package_name: str
    version_name: Optional[str]
    version_code: Optional[int]
    flags: Union[str, list]
    first_install_time: datetime.datetime
    last_update_time: datetime.datetime
    signature: str
    path: str
    sub_apk_paths: List[str]


@dataclass
class BatteryInfo:
    ac_powered: bool
    usb_powered: bool
    wireless_powered: Optional[bool]
    dock_powered: Optional[bool]
    max_charging_current: Optional[int]
    max_charging_voltage: Optional[int]
    charge_counter: Optional[int]
    status: Optional[int]
    health: Optional[int]
    present: Optional[bool]
    level: Optional[int]
    scale: Optional[int]
    voltage: Optional[int]  # mV
    temperature: Optional[float]  # e.g. 25.0
    technology: Optional[str]


class WindowSize(NamedTuple):
    width: int
    height: int


@dataclass
class RunningAppInfo:
    package: str
    activity: str
    pid: int = 0


@dataclass
class ShellReturn:
    """
    Attributes:
        command: The str command passed to run().
        returncode: The exit code of the process, negative for signals.
        output: the output
    """
    command: str
    returncode: int
    output: str | bytes


@dataclass
class AdbDeviceInfo:
    serial: str
    state: str


StrOrPathLike = Union[str, pathlib.Path]
